@@ -1,25 +1,7 @@
-###############################################################################
-#
-# Copyright (c) 2009-2017 Cedric Krier.
-# Copyright (c) 2007, 2008 OpenHex SPRL. (http://openhex.com) All Rights
-# Reserved.
-#
-# This program is free software; you can redistribute it and/or modify it under
-# the terms of the GNU General Public License as published by the Free Software
-# Foundation; either version 3 of the License, or (at your option) any later
-# version.
-#
-# This program is distributed in the hope that it will be useful, but WITHOUT
-# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
-# FOR A PARTICULAR PURPOSE. See the GNU General Public License for more
-# details.
-#
-# You should have received a copy of the GNU General Public License along with
-# this program.  If not, see <http://www.gnu.org/licenses/>.
-#
-###############################################################################
-
+# This file is part of relatorio.  The COPYRIGHT file at the top level of
+# this repository contains the full copyright notices and license terms.
 import re
+
 try:
     # requires python 2.5+
     from hashlib import md5
@@ -27,44 +9,34 @@ except ImportError:
     from md5 import md5
 
 import base64
+import datetime
 import mimetypes
 import sys
 import time
-import urllib.request, urllib.parse, urllib.error
-import zipfile
-from io import BytesIO
-from copy import deepcopy
-import datetime
-from decimal import Decimal
-
-
+import urllib.parse
 import warnings
+import zipfile
+from copy import deepcopy
+from decimal import Decimal
+from io import BytesIO
 
-import lxml.etree
 import genshi
 import genshi.output
-from genshi.template import MarkupTemplate
+import lxml.etree
+from genshi.core import Stream
 from genshi.filters import Transformer
 from genshi.filters.transform import ENTER, EXIT
-from genshi.core import Stream
+from genshi.template import MarkupTemplate
 from genshi.template.interpolation import PREFIX
 
 import relatorio
+from relatorio.reporting import MIMETemplateLoader, Report
 from relatorio.templates.base import RelatorioStream
-from relatorio.reporting import Report, MIMETemplateLoader
+
 try:
     from relatorio.templates.chart import Template as ChartTemplate
 except ImportError:
-    ChartTemplate = type(None)
-
-try:
-    str
-except NameError:
-    str = str
-try:
-    int
-except NameError:
-    long = int
+    ChartTemplate = None
 
 __metaclass__ = type
 warnings.filterwarnings('always', module='relatorio.templates.opendocument')
@@ -132,15 +104,18 @@ class ImageHref:
         bitstream, mimetype = expr[:2]
         if isinstance(bitstream, Report):
             bitstream = bitstream(**self.context).render()
-        elif isinstance(bitstream, ChartTemplate):
+        elif ChartTemplate and isinstance(bitstream, ChartTemplate):
             bitstream = bitstream.generate(**self.context).render()
         elif not hasattr(bitstream, 'seek') or not hasattr(bitstream, 'read'):
             bitstream = BytesIO(bitstream)
-        bitstream.seek(0)
-        file_content = bitstream.read()
+        if bitstream:
+            bitstream.seek(0)
+            file_content = bitstream.read()
+        else:
+            file_content = b''
         name = md5(file_content).hexdigest()
         path = 'Pictures/%s%s' % (
-            name, mimetypes.guess_extension(mimetype))
+            name, mimetypes.guess_extension(mimetype or '') or '')
         self.serializer.add_file(path, file_content, mimetype)
         return {'{http://www.w3.org/1999/xlink}href': path}
 
@@ -267,7 +242,16 @@ class Template(MarkupTemplate):
 
     def __init__(self, source, filepath=None, filename=None, loader=None,
                  encoding=None, lookup='strict', allow_exec=True):
-        self.namespaces = {}
+        # assign default/fake namespaces so that documents do not need to
+        # define them if they don't use them
+        self.namespaces = {
+            "text": "urn:text",
+            "draw": "urn:draw",
+            "table": "urn:table",
+            "office": "urn:office",
+            "xlink": "urn:xlink",
+            "svg": "urn:svg",
+        }
         self.inner_docs = []
         self.has_col_loop = False
         self._source = None
@@ -297,29 +281,37 @@ class Template(MarkupTemplate):
         zf = get_zip_file(source)
         content = zf.read('content.xml')
         styles = zf.read('styles.xml')
+        meta = zf.read('meta.xml')
 
         template = super(Template, self)
         content = template._parse(self.insert_directives(content), encoding)
         styles = template._parse(self.insert_directives(styles), encoding)
+        meta = template._parse(self.insert_directives(meta), encoding)
         content_files = [('content.xml', content)]
         styles_files = [('styles.xml', styles)]
+        meta_files = [('meta.xml', meta)]
 
         while self.inner_docs:
             doc = self.inner_docs.pop()
-            c_path, s_path = doc + '/content.xml', doc + '/styles.xml'
+            c_path, s_path, m_path = (
+                doc + '/content.xml', doc + '/styles.xml', doc + '/meta.xml')
             content = zf.read(c_path)
             styles = zf.read(s_path)
+            meta = zf.read(m_path)
 
             c_parsed = template._parse(self.insert_directives(content),
                                        encoding)
             s_parsed = template._parse(self.insert_directives(styles),
                                        encoding)
+            m_parsed = template._parse(self.insert_directives(meta),
+                                       encoding)
             content_files.append((c_path, c_parsed))
             styles_files.append((s_path, s_parsed))
+            meta_files.append((m_path, m_parsed))
         zf.close()
 
         parsed = []
-        for fpath, fparsed in content_files + styles_files:
+        for fpath, fparsed in content_files + styles_files + meta_files:
             self._files.add(fpath)
             parsed.append((genshi.core.PI, ('relatorio', fpath), None))
             parsed += fparsed
@@ -332,16 +324,6 @@ class Template(MarkupTemplate):
         tree = lxml.etree.parse(BytesIO(content))
         root = tree.getroot()
 
-        # assign default/fake namespaces so that documents do not need to
-        # define them if they don't use them
-        self.namespaces = {
-            "text": "urn:text",
-            "draw": "urn:draw",
-            "table": "urn:table",
-            "office": "urn:office",
-            "xlink": "urn:xlink",
-            "svg": "urn:svg",
-        }
         # but override them with the real namespaces
         self.namespaces.update(root.nsmap)
 
@@ -353,6 +335,7 @@ class Template(MarkupTemplate):
 
         self._remove_soft_page_break(tree)
         self._invert_style(tree)
+        self._handle_meta(tree)
         self._handle_relatorio_tags(tree)
         self._handle_images(tree)
         self._handle_innerdocs(tree)
@@ -402,7 +385,8 @@ class Template(MarkupTemplate):
             if statement.tag == placeholder:
                 expr = statement.text[1:-1]
             elif statement.tag == text_a:
-                expr = urllib.parse.unquote(statement.attrib[xlink_href_attrib][12:])
+                expr = urllib.parse.unquote(
+                    statement.attrib[xlink_href_attrib][12:])
 
             if not expr:
                 raise OOTemplateError("No expression in the tag",
@@ -415,7 +399,9 @@ class Template(MarkupTemplate):
             if not statement.text:
                 warn_msg = "No statement text in '%s' for '%s'" \
                            % (self.filepath, expr)
-            elif expr != statement.text and statement.tag == text_a:
+            elif (directive != 'attrs'
+                    and expr != statement.text
+                    and statement.tag == text_a):
                 warn_msg = "url and text do not match in %s: %s != %s" \
                            % (self.filepath, expr, statement.text)
             if warn_msg:
@@ -443,6 +429,45 @@ class Template(MarkupTemplate):
         assert not opened_tags
         return r_statements, closing_tags
 
+    def _handle_meta(self, tree):
+        """updates meta
+        and adds py:content into meta:user-defined and dc:* nodes"""
+        root = tree.getroot()
+        if root.tag != '{%s}document-meta' % self.namespaces['office']:
+            return
+        xpath_expr = ("//meta:user-defined[starts-with(., 'relatorio://')]"
+                      "|//dc:*[starts-with(., 'relatorio://')]")
+        genshi_content = '{%s}content' % self.namespaces['py']
+        for node in tree.xpath(xpath_expr, namespaces=self.namespaces):
+            node.attrib[genshi_content] = node.text[len('relatorio://'):]
+
+        def set(name, value):
+            meta = root.find('{%s}%s' % (self.namespaces['meta'], name))
+            if meta is None:
+                meta = EtreeElement(
+                    '{%s}%s' % (self.namespaces['meta'], name),
+                    nsmap=self.namespaces)
+                root.append(meta)
+            meta.text = value
+
+        def remove(name, namespace='meta'):
+            meta = tree.find('{%s}%s' % (self.namespaces[namespace], name))
+            if meta is not None:
+                tree.remove(meta)
+
+        now = datetime.datetime.now()
+        set('creation-date', now.isoformat())
+        set('date', now.isoformat())
+        remove('document-statistic')
+        set('editing-cycles', '1')
+        remove('editing-duration')
+        set('generator', 'relatorio/%s' % relatorio.__version__)
+        remove('initial-creator')
+        remove('print-date')
+        remove('printed-by')
+        remove('creator', 'dc')
+        remove('date', 'dc')
+
     def _handle_relatorio_tags(self, tree):
         """
         Will treat all relatorio tag (py:if/for/choose/when/otherwise)
@@ -537,7 +562,8 @@ class Template(MarkupTemplate):
                 parent.attrib[py_attr] = a_val
 
                 # remove the directive node
-                r_node.getparent().remove(r_node)
+                if parent != r_node:
+                    remove_node_keeping_tail(r_node)
             else:
                 def has_style(node):
                     return any(attr in node.attrib
@@ -586,7 +612,8 @@ class Template(MarkupTemplate):
         repeat_tag = '{%s}repeat' % RELATORIO_URI
 
         # table node (it is not necessarily the direct parent of ancestor)
-        table_node = next(ancestor.iterancestors('{%s}table' % table_namespace))
+        table_node = next(
+            ancestor.iterancestors('{%s}table' % table_namespace))
         table_name = table_node.attrib['{%s}name' % table_namespace]
 
         # add counting instructions
@@ -1011,51 +1038,6 @@ class Manifest(object):
             self.root.remove(entry)
 
 
-class Meta(object):
-
-    def __init__(self, content):
-        self.tree = lxml.etree.parse(BytesIO(content))
-        root = self.tree.getroot()
-        self.namespaces = root.nsmap
-        path = '/office:document-meta/office:meta'
-        self.office_meta, = self.tree.xpath(path, namespaces=self.namespaces)
-
-    def set(self, name, value, namespace='meta'):
-        namespace = self.namespaces[namespace]
-        meta = self.office_meta.find('{%s}%s' % (namespace, name))
-        if meta is None:
-            meta = EtreeElement('{%s}%s' % (namespace, name),
-                                nsmap={'meta': namespace})
-            self.office_meta.append(meta)
-        meta.text = value
-
-    def remove(self, name, namespace='meta'):
-        namespace = self.namespaces[namespace]
-        meta = self.office_meta.find('{%s}%s' % (namespace, name))
-        if meta is not None:
-            self.office_meta.remove(meta)
-
-    def __str__(self):
-        now = datetime.datetime.now()
-        self.set('creation-date', now.isoformat())
-        self.set('date', now.isoformat())
-        self.remove('document-statistic')
-        self.set('editing-cycles', '1')
-        self.remove('editing-duration')
-        self.set('generator', 'relatorio/%s' % relatorio.__version__)
-        self.remove('initial-creator')
-        self.remove('print-date')
-        self.remove('printed-by')
-        self.remove('creator', 'dc')
-        self.remove('date', 'dc')
-        val = lxml.etree.tostring(self.tree, encoding='UTF-8',
-                                  xml_declaration=True)
-        # In Python 3, val will be bytes
-        if not isinstance(val, str):
-            return str(val, 'utf-8')
-        return val
-
-
 class _AbstractZipWriteSplitStream(object):
     def __init__(self, zipfile, chunksize=64):
         self.zipfile = zipfile
@@ -1137,12 +1119,11 @@ class OOSerializer:
     def __init__(self, source, files, chunksize=64):
         self.inzip = get_zip_file(source)
         self.manifest = Manifest(self.inzip.read(MANIFEST))
-        self.meta = Meta(self.inzip.read(META))
         self.xml_serializer = genshi.output.XMLSerializer()
         self._files = files
         self.chunksize = chunksize
         self.outzip = None
-        self._deferred = None
+        self._deferred = []
 
     def __call__(self, stream, method=None, encoding='utf-8', out=None):
         if out is None:
@@ -1151,7 +1132,6 @@ class OOSerializer:
             result = out
         self.outzip = zipfile.ZipFile(
             result, mode='w', compression=zipfile.ZIP_DEFLATED)
-        self._deferred = []
         files = {}
         now = time.localtime()[:6]
         manifest_info = None
@@ -1167,8 +1147,6 @@ class OOSerializer:
                 files[f_info.filename] = new_info
             elif f_info.filename == MANIFEST:
                 manifest_info = f_info
-            elif f_info.filename == META:
-                self.outzip.writestr(f_info, str(self.meta))
             elif f_info.filename.startswith(THUMBNAILS + '/'):
                 self.manifest.remove_file_entry(f_info.filename)
             else:
@@ -1190,7 +1168,9 @@ class OOSerializer:
             return result
 
     def add_file(self, path, content, mimetype):
-        if self.outzip and path not in self.outzip.namelist():
+        if not self.outzip:
+            self._deferred.append((path, content, mimetype))
+        elif path not in self.outzip.namelist():
             try:
                 self.outzip.writestr(path, content)
                 self.manifest.add_file_entry(path, mimetype)
