@@ -1,43 +1,18 @@
 # -*- encoding: utf-8 -*-
-###############################################################################
-#
-# Copyright (c) 2014-2017 Cedric Krier.
-# Copyright (c) 2007, 2008 OpenHex SPRL. (http://openhex.com) All Rights
-# Reserved.
-#
-# This program is free software; you can redistribute it and/or modify it under
-# the terms of the GNU General Public License as published by the Free Software
-# Foundation; either version 3 of the License, or (at your option) any later
-# version.
-#
-# This program is distributed in the hope that it will be useful, but WITHOUT
-# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
-# FOR A PARTICULAR PURPOSE. See the GNU General Public License for more
-# details.
-#
-# You should have received a copy of the GNU General Public License along with
-# this program.  If not, see <http://www.gnu.org/licenses/>.
-#
-###############################################################################
-
-
+# This file is part of relatorio.  The COPYRIGHT file at the top level of
+# this repository contains the full copyright notices and license terms.
 import os
 import unittest
-from io import StringIO, BytesIO
+from io import BytesIO, StringIO
 
 import lxml.etree
-from genshi.filters import Translator
 from genshi.core import PI
+from genshi.filters import Translator
 from genshi.template.eval import UndefinedError
 
-from relatorio.templates.opendocument import Template, GENSHI_EXPR,\
-    GENSHI_URI, RELATORIO_URI, fod2od, remove_node_keeping_tail, \
-    escape_xml_invalid_chars
-
-try:
-    str
-except NameError:
-    str = str
+from relatorio.templates.opendocument import (
+    GENSHI_EXPR, GENSHI_URI, RELATORIO_URI, Template, escape_xml_invalid_chars,
+    fod2od, remove_node_keeping_tail)
 
 OO_TABLE_NS = "urn:oasis:names:tc:opendocument:xmlns:table:1.0"
 
@@ -79,14 +54,19 @@ class TestOOTemplating(unittest.TestCase):
                      'images': [(open(os.path.join(thisdir, 'one.jpg'), 'rb'),
                                  'image/jpeg', None, None, 'One'),
                                 (open(os.path.join(thisdir, 'two.png'), 'rb'),
-                                 'image/png', '2cm', '2.2cm', 'Two')],
+                                 'image/png', '2cm', '2.2cm', 'Two'),
+                                (None, None)],
                      'oeuf': open(os.path.join(thisdir, 'egg.jpg'), 'rb'),
-                     'footer': 'We sell stuff'}
+                     'footer': 'We sell stuff',
+                     'salutation': 'Greating',
+                     'title': 'Testing Letter',
+                     }
 
     def tearDown(self):
         self._source.close()
         for image in self.data['images']:
-            image[0].close()
+            if image[0]:
+                image[0].close()
         self.data['oeuf'].close()
 
     def test_init(self):
@@ -120,7 +100,7 @@ class TestOOTemplating(unittest.TestCase):
                             <text:a xlink:href="relatorio://foo">foo</text:a>
                         </text:p>
                     </table:table-cell>
-                 </xml>'''
+                 </xml>'''  # noqa: E501
         interpolated = self.oot.insert_directives(xml)
         root_interpolated = lxml.etree.parse(interpolated).getroot()
         cell = root_interpolated[0]
@@ -142,7 +122,7 @@ class TestOOTemplating(unittest.TestCase):
                         </text:p>
                         <text:p>bar</text:p>
                     </table:table-cell>
-                 </xml>'''
+                 </xml>'''  # noqa: E501
         interpolated = self.oot.insert_directives(xml)
         root_interpolated = lxml.etree.parse(interpolated).getroot()
         cell = root_interpolated[0]
@@ -163,7 +143,7 @@ class TestOOTemplating(unittest.TestCase):
                             <text:a xlink:href="relatorio://foo">foo</text:a>
                         </text:p>
                     </table:table-cell>
-                 </xml>'''
+                 </xml>'''  # noqa: E501
         interpolated = self.oot.insert_directives(xml)
         root_interpolated = lxml.etree.parse(interpolated).getroot()
         cell = root_interpolated[0]
@@ -184,7 +164,7 @@ class TestOOTemplating(unittest.TestCase):
                             bar
                         </text:p>
                     </table:table-cell>
-                 </xml>'''
+                 </xml>'''  # noqa: E501
         interpolated = self.oot.insert_directives(xml)
         root_interpolated = lxml.etree.parse(interpolated).getroot()
         cell = root_interpolated[0]
@@ -196,6 +176,35 @@ class TestOOTemplating(unittest.TestCase):
             child.get('{http://genshi.edgewall.org/}replace'),
             '__relatorio_escape_invalid_chars(foo)')
 
+    def test_directives_with_tail(self):
+        "Testing directives with tail"
+        xml = b'''<xml xmlns:text="urn:text">
+                    <text:p>before
+                        <text:placeholder>&lt;attrs text:p=&quot;{}&quot;&gt;</text:placeholder>
+                        after
+                    </text:p>
+                </xml>'''  # noqa: E501
+        interpolated = self.oot.insert_directives(xml)
+        root_interpolated = lxml.etree.parse(interpolated).getroot()
+        paragraph = root_interpolated[0]
+        self.assertEqual(len(paragraph), 0)
+        self.assertEqual(
+            ''.join(map(str.strip, paragraph.text.split())), 'beforeafter')
+
+    def test_directives_on_self(self):
+        "Testing directive that applies on itself"
+        xml = b'''<xml xmlns:text="urn:text" xmlns:xlink="urn:xlink">
+                    <text:a xlink:href="relatorio://attrs text:a=&quot;{'{urn:xlink}href': 'foo'}&quot;">
+                    test
+                    </text:a>
+                </xml>'''  # noqa: E501
+        interpolated = self.oot.insert_directives(xml)
+        root_interpolated = lxml.etree.parse(interpolated).getroot()
+        child = root_interpolated[0]
+        self.assertEqual(
+            child.get('{http://genshi.edgewall.org/}attrs'),
+            "{'{urn:xlink}href': 'foo'}")
+
     def test_column_looping(self):
         xml = b'''
 <table:table
@@ -394,6 +403,13 @@ class TestOOTemplating(unittest.TestCase):
         self.assertRaises(UndefinedError,
             lambda: stream.events.render(encoding='utf-8'))
 
+    def test_meta(self):
+        "Testing that meta get rendered"
+        stream = self.oot.generate(**self.data)
+        rendered = stream_to_string(stream.events.render(encoding='utf-8'))
+        self.assertTrue("Greating" in rendered)
+        self.assertTrue("Testing Letter" in rendered)
+
     def test_generate(self):
         "Testing that content get rendered"
         stream = self.oot.generate(**self.data)
@@ -425,7 +441,7 @@ class TestOOTemplating(unittest.TestCase):
         tree = lxml.etree.parse(StringIO(rendered[25:styles_idx]))
         root = tree.getroot()
         images = root.xpath('//draw:frame', namespaces=self.oot.namespaces)
-        self.assertEqual(len(images), 4)
+        self.assertEqual(len(images), 5)
         self.assertFalse(
             images[0].get('{%s}name' % self.oot.namespaces['draw']))
         self.assertFalse(
